@@ -1,6 +1,5 @@
-from .client import Client  # noqa: F401
+### legacy DNS bindings - DEPRECATED ###
 
-from .dns_providers import BaseDns  # noqa: F401
 from .dns_providers import AuroraDns  # noqa: F401
 from .dns_providers import CloudFlareDns  # noqa: F401
 from .dns_providers import AcmeDnsDns  # noqa: F401
@@ -11,3 +10,5 @@ from .dns_providers import DNSPodDns  # noqa:F401
 from .dns_providers import DuckDNSDns  # noqa:F401
 from .dns_providers import ClouDNSDns  # noqa:F401
 from .dns_providers import Route53Dns  # noqa:F401
+from .dns_providers import PowerDNSDns  # noqa:F401
+from .dns_providers import GandiDns  # noqa:F401
