@@ -1,10 +1,10 @@
 import os
-import logging
 import argparse
 
-from . import Client
+from .client import Client
 from . import __version__ as sewer_version
 from .config import ACME_DIRECTORY_URL_STAGING, ACME_DIRECTORY_URL_PRODUCTION
+from .lib import create_logger
 
 
 def main():
@@ -72,6 +72,8 @@ def main():
             "dnspod",
             "duckdns",
             "cloudns",
+            "powerdns",
+            "gandi",
         ],
         help="The name of the dns provider that you want to use.",
     )
@@ -165,13 +167,7 @@ def main():
     if not os.access(out_dir, os.W_OK):
         raise OSError("The dir '{0}' is not writable".format(out_dir))
 
-    logger = logging.getLogger()
-    handler = logging.StreamHandler()
-    formatter = logging.Formatter("%(message)s")
-    handler.setFormatter(formatter)
-    if not logger.handlers:
-        logger.addHandler(handler)
-    logger.setLevel(loglevel)
+    logger = create_logger(None, loglevel)
 
     if account_key:
         account_key = account_key.read()
@@ -306,6 +302,29 @@ def main():
         except KeyError as e:
             logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
             raise
+    elif dns_provider == "powerdns":
+        from . import PowerDNSDns
+
+        try:
+            powerdns_api_key = os.environ["POWERDNS_API_KEY"]
+            powerdns_api_url = os.environ["POWERDNS_API_URL"]
+
+            dns_class = PowerDNSDns(powerdns_api_key, powerdns_api_url)
+            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
+        except KeyError as e:
+            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
+            raise
+    elif dns_provider == "gandi":
+        from . import GandiDns
+
+        try:
+            gandi_api_key = os.environ["GANDI_API_KEY"]
+
+            dns_class = GandiDns(GANDI_API_KEY=gandi_api_key)
+            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
+        except KeyError as e:
+            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
+            raise
     else:
         raise ValueError("The dns provider {0} is not recognised.".format(dns_provider))
 
