@@ -1,11 +1,10 @@
 import time
 import copy
 import json
-import base64
-import hashlib
-import logging
+from hashlib import sha256
 import binascii
 import platform
+from typing import Dict, Union, cast
 
 import requests
 import OpenSSL
@@ -13,20 +12,24 @@ import cryptography
 
 from . import __version__ as sewer_version
 from .config import ACME_DIRECTORY_URL_PRODUCTION
+from .lib import create_logger, log_response, safe_base64
 
 
-class Client(object):
+class Client:
     """
     todo: improve documentation.
 
     usage:
         import sewer
-        dns_class = sewer.CloudFlareDns(CLOUDFLARE_EMAIL='example@example.com',
-                                        CLOUDFLARE_API_KEY='nsa-grade-api-key')
+        provider = sewer.CloudFlareDns(
+            CLOUDFLARE_EMAIL='example@example.com',
+            CLOUDFLARE_API_KEY='nsa-grade-api-key'
+        )
 
-        1. to create a new certificate.
-        client = sewer.Client(domain_name='example.com',
-                              dns_class=dns_class)
+        ### to create a new certificate.
+        client = sewer.Client(
+            domain_name='example.com', provider=provider
+        )
         certificate = client.cert()
         certificate_key = client.certificate_key
         account_key = client.account_key
@@ -37,42 +40,47 @@ class Client(object):
         with open('certificate.key', 'w') as certificate_key_file:
             certificate_key_file.write(certificate_key)
 
-
-        2. to renew a certificate:
+        ### to renew a certificate:
         with open('account_key.key', 'r') as account_key_file:
             account_key = account_key_file.read()
 
-        client = sewer.Client(domain_name='example.com',
-                              dns_class=dns_class,
-                              account_key=account_key)
+        client = sewer.Client(
+            domain_name='example.com', provider=provider,
+            account_key=account_key
+        )
         certificate = client.renew()
         certificate_key = client.certificate_key
 
-    todo:
-        - handle more exceptions
+    todo: handle more exceptions
     """
 
     def __init__(
         self,
         domain_name,
-        dns_class,
+        dns_class=None,
         domain_alt_names=None,
         contact_email=None,
         account_key=None,
         certificate_key=None,
         bits=2048,
         digest="sha256",
+        provider=None,
         ACME_REQUEST_TIMEOUT=7,
         ACME_AUTH_STATUS_WAIT_PERIOD=8,
         ACME_AUTH_STATUS_MAX_CHECKS=3,
         ACME_DIRECTORY_URL=ACME_DIRECTORY_URL_PRODUCTION,
+        ACME_VERIFY=True,
         LOG_LEVEL="INFO",
     ):
         """
         :param domain_name:                  (required) [string]
             the name that you want to acquire/renew certificate for. wildcards are allowed.
         :param dns_class:                    (required) [class]
-            a subclass of sewer.BaseDns which will be called to create/delete DNS TXT records.
+            (DEPRECATED) a subclass of BaseDns which will be called to create/delete DNS TXT records.
+            do not pass this parameter if also passing provider.
+        :param provider:                (required) [class]
+            a subclass of ProviderBase which will be called to create/delete auth records.
+            do not pass this parameter if also passing dns_class
         :param domain_alt_names:             (optional) [list]
             list of alternative names that you want to be bundled into the same certificate as domain_name.
         :param contact_email:                (optional) [string]
@@ -95,6 +103,8 @@ class Client(object):
             the max number of times the client will poll the acme server to check on authorization status
         :param ACME_DIRECTORY_URL:           (optional) [string]
             the url of the acme servers' directory endpoint
+        :param ACME_VERIFY:                  (optional) [bool]
+            suppress verification of SSL cert when set to False (for pebble); hint: -Wignore
         :param LOG_LEVEL:                    (optional) [string]
             the level to output log messages at. one of; 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'
         """
@@ -131,9 +141,13 @@ class Client(object):
                     LOG_LEVEL
                 )
             )
+        elif dns_class is not None and provider is not None:
+            raise ValueError(
+                "passed both the DEPRECATED 'dns_class' parameter as well as 'provider'."
+            )
 
         self.domain_name = domain_name
-        self.dns_class = dns_class
+        self.provider = provider if provider is not None else dns_class
         if not domain_alt_names:
             domain_alt_names = []
         self.domain_alt_names = domain_alt_names
@@ -145,15 +159,10 @@ class Client(object):
         self.ACME_AUTH_STATUS_WAIT_PERIOD = ACME_AUTH_STATUS_WAIT_PERIOD
         self.ACME_AUTH_STATUS_MAX_CHECKS = ACME_AUTH_STATUS_MAX_CHECKS
         self.ACME_DIRECTORY_URL = ACME_DIRECTORY_URL
+        self.ACME_VERIFY = ACME_VERIFY
         self.LOG_LEVEL = LOG_LEVEL.upper()
 
-        self.logger = logging.getLogger(__name__)
-        self.logger.setLevel(self.LOG_LEVEL)
-        if not self.logger.hasHandlers():
-            handler = logging.StreamHandler()
-            formatter = logging.Formatter("%(message)s")
-            handler.setFormatter(formatter)
-            self.logger.addHandler(handler)
+        self.logger = create_logger(__name__, LOG_LEVEL)
 
         try:
             self.all_domain_names = copy.copy(self.domain_alt_names)
@@ -183,6 +192,11 @@ class Client(object):
                 self.account_key = account_key
                 self.PRIOR_REGISTERED = True
 
+            if dns_class is not None:
+                self.logger.warning(
+                    "DEPRECATED parameter 'dns_class' will be removed in 0.9; use 'provider' instead"
+                )
+
             self.logger.info(
                 "intialise_success, sewer_version={0}, domain_names={1}, acme_server={2}".format(
                     sewer_version.__version__,
@@ -190,21 +204,66 @@ class Client(object):
                     self.ACME_DIRECTORY_URL[:20] + "...",
                 )
             )
+
+        ### FIX ME ### [:100] is bandaid to reduce spew during tests
+
         except Exception as e:
-            self.logger.error("Unable to intialise client. error={0}".format(str(e)))
+            self.logger.error("Unable to intialise Client. error={0}".format(str(e)[:100]))
             raise e
 
-    @staticmethod
-    def log_response(response):
+    def GET(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
         """
-        renders response as json or as a string
+        wrap requests.get (and post and head, below) to allow:
+          * injection of e.g. UserAgent header in one place rather than all over
+          * hides requests itself to allow for change (unlikely) or use of Session
+          * paves the way to inject the verify option, required to use pebble
         """
-        # TODO: use this to handle all response logs.
-        try:
-            log_body = response.json()
-        except ValueError:
-            log_body = response.content[:30]
-        return log_body
+
+        return self._request("GET", url, **kwargs)
+
+    def HEAD(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
+        return self._request("HEAD", url, **kwargs)
+
+    def POST(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
+        return self._request("POST", url, **kwargs)
+
+    def _request(
+        self, method: str, url: str, **kwargs: Union[str, Dict[str, str], bool]
+    ) -> requests.Response:
+        """
+        shared implementation for GET, POST and HEAD
+        * injects standard request options unless they are already given in kwargs
+          * header:UserAgent, timeout
+          * verify - this is a hack to make sewer accept pebble's intentionally bogus cert
+        """
+
+        # if there's no UserAgent header in args, inject it
+        headers = cast(Dict[str, str], kwargs.setdefault("headers", {}))
+        if "UserAgent" not in headers:
+            headers["UserAgent"] = self.User_Agent
+
+        # add standard timeout if there's none already present
+        if "timeout" not in kwargs:
+            kwargs["timeout"] = self.ACME_REQUEST_TIMEOUT
+
+        ### FIX ME ### can get current bogus cert from pebble, figure out how to use it here?
+
+        # if ACME_VERIFY is false, disable certificate check in request
+        if not self.ACME_VERIFY:
+            kwargs["verify"] = False
+
+        # awkward implementation to maintain compatibility with current mocked tests
+        if method == "GET":
+            ### Incredibly stupid hack to reconcile mypy to requests' API
+            response = requests.get(url, None, **kwargs)
+        elif method == "HEAD":
+            response = requests.head(url, **kwargs)
+        elif method == "POST":
+            ### Incredibly stupid hack to reconcile mypy to requests' API
+            data_hack = cast(Union[Dict[str, str], None], kwargs.pop("data", None))
+            response = requests.post(url, data_hack, **kwargs)
+
+        return response
 
     @staticmethod
     def get_user_agent():
@@ -218,10 +277,7 @@ class Client(object):
 
     def get_acme_endpoints(self):
         self.logger.debug("get_acme_endpoints")
-        headers = {"User-Agent": self.User_Agent}
-        get_acme_endpoints = requests.get(
-            self.ACME_DIRECTORY_URL, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers
-        )
+        get_acme_endpoints = self.GET(self.ACME_DIRECTORY_URL)
         self.logger.debug(
             "get_acme_endpoints_response. status_code={0}".format(get_acme_endpoints.status_code)
         )
@@ -229,7 +285,7 @@ class Client(object):
             raise ValueError(
                 "Error while getting Acme endpoints: status_code={status_code} response={response}".format(
                     status_code=get_acme_endpoints.status_code,
-                    response=self.log_response(get_acme_endpoints),
+                    response=log_response(get_acme_endpoints),
                 )
             )
         return get_acme_endpoints
@@ -296,7 +352,7 @@ class Client(object):
         provided, then the server SHOULD use status code 409 (Conflict) and
         provide the URL of that account in the Location header field
         """
-        self.logger.info("acme_register")
+        self.logger.info("acme_register (newAccount)")
         if self.PRIOR_REGISTERED:
             payload = {"onlyReturnExisting": True}
         elif self.contact_email:
@@ -308,10 +364,12 @@ class Client(object):
             payload = {"termsOfServiceAgreed": True}
 
         url = self.ACME_NEW_ACCOUNT_URL
-        acme_register_response = self.make_signed_acme_request(url=url, payload=payload)
+        acme_register_response = self.make_signed_acme_request(
+            url=url, payload=json.dumps(payload), needs_jwk=True
+        )
         self.logger.debug(
             "acme_register_response. status_code={0}. response={1}".format(
-                acme_register_response.status_code, self.log_response(acme_register_response)
+                acme_register_response.status_code, log_response(acme_register_response)
             )
         )
 
@@ -319,7 +377,7 @@ class Client(object):
             raise ValueError(
                 "Error while registering: status_code={status_code} response={response}".format(
                     status_code=acme_register_response.status_code,
-                    response=self.log_response(acme_register_response),
+                    response=log_response(acme_register_response),
                 )
             )
 
@@ -347,18 +405,20 @@ class Client(object):
         The date values seem to be ignored by LetsEncrypt although they are
         in the ACME draft spec; https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.4
         """
-        self.logger.info("apply_for_cert_issuance")
+        self.logger.info("apply_for_cert_issuance (newOrder)")
         identifiers = []
         for domain_name in self.all_domain_names:
             identifiers.append({"type": "dns", "value": domain_name})
 
         payload = {"identifiers": identifiers}
         url = self.ACME_NEW_ORDER_URL
-        apply_for_cert_issuance_response = self.make_signed_acme_request(url=url, payload=payload)
+        apply_for_cert_issuance_response = self.make_signed_acme_request(
+            url=url, payload=json.dumps(payload)
+        )
         self.logger.debug(
             "apply_for_cert_issuance_response. status_code={0}. response={1}".format(
                 apply_for_cert_issuance_response.status_code,
-                self.log_response(apply_for_cert_issuance_response),
+                log_response(apply_for_cert_issuance_response),
             )
         )
 
@@ -366,7 +426,7 @@ class Client(object):
             raise ValueError(
                 "Error applying for certificate issuance: status_code={status_code} response={response}".format(
                     status_code=apply_for_cert_issuance_response.status_code,
-                    response=self.log_response(apply_for_cert_issuance_response),
+                    response=log_response(apply_for_cert_issuance_response),
                 )
             )
 
@@ -377,7 +437,7 @@ class Client(object):
         self.logger.info("apply_for_cert_issuance_success")
         return authorizations, finalize_url
 
-    def get_identifier_authorization(self, url):
+    def get_identifier_authorization(self, auth_url: str) -> Dict[str, str]:
         """
         https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.5
         When a client receives an order from the server it downloads the
@@ -388,63 +448,53 @@ class Client(object):
 
         This is also where we get the challenges/tokens.
         """
-        self.logger.info("get_identifier_authorization")
-        headers = {"User-Agent": self.User_Agent}
-        get_identifier_authorization_response = requests.get(
-            url, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers
-        )
+        self.logger.info("get_identifier_authorization for %s" % auth_url)
+        response = self.make_signed_acme_request(auth_url, payload="")
+
         self.logger.debug(
             "get_identifier_authorization_response. status_code={0}. response={1}".format(
-                get_identifier_authorization_response.status_code,
-                self.log_response(get_identifier_authorization_response),
+                response.status_code, log_response(response)
             )
         )
-        if get_identifier_authorization_response.status_code not in [200, 201]:
+        if response.status_code not in [200, 201]:
             raise ValueError(
                 "Error getting identifier authorization: status_code={status_code} response={response}".format(
-                    status_code=get_identifier_authorization_response.status_code,
-                    response=self.log_response(get_identifier_authorization_response),
+                    status_code=response.status_code, response=log_response(response)
                 )
             )
-        res = get_identifier_authorization_response.json()
-        domain = res["identifier"]["value"]
-        wildcard = res.get("wildcard")
-        if wildcard:
-            domain = "*." + domain
+        response_json = response.json()
+        domain = response_json["identifier"]["value"]
+        wildcard = response_json.get("wildcard")
 
-        for i in res["challenges"]:
-            if i["type"] == "dns-01":
-                dns_challenge = i
-        dns_token = dns_challenge["token"]
-        dns_challenge_url = dns_challenge["url"]
-        identifier_auth = {
-            "domain": domain,
-            "url": url,
-            "wildcard": wildcard,
-            "dns_token": dns_token,
-            "dns_challenge_url": dns_challenge_url,
-        }
+        for i in response_json["challenges"]:
+            if i["type"] in self.provider.chal_types:
+                challenge = i
+                challenge_token = challenge["token"]
+                challenge_url = challenge["url"]
+
+                identifier_auth = {
+                    "domain": domain,
+                    "url": auth_url,
+                    "wildcard": wildcard,
+                    "token": challenge_token,
+                    "challenge_url": challenge_url,
+                }
 
         self.logger.debug(
             "get_identifier_authorization_success. identifier_auth={0}".format(identifier_auth)
         )
-        self.logger.info("get_identifier_authorization_success")
+        self.logger.info(
+            "get_identifier_authorization got %s, token=%s" % (challenge_url, challenge_token)
+        )
         return identifier_auth
 
-    def get_keyauthorization(self, dns_token):
+    def get_keyauthorization(self, token):
         self.logger.debug("get_keyauthorization")
-        acme_header_jwk_json = json.dumps(
-            self.get_acme_header("GET_THUMBPRINT")["jwk"], sort_keys=True, separators=(",", ":")
-        )
-        acme_thumbprint = self.calculate_safe_base64(
-            hashlib.sha256(acme_header_jwk_json.encode("utf8")).digest()
-        )
-        acme_keyauthorization = "{0}.{1}".format(dns_token, acme_thumbprint)
-        base64_of_acme_keyauthorization = self.calculate_safe_base64(
-            hashlib.sha256(acme_keyauthorization.encode("utf8")).digest()
-        )
+        acme_header_jwk_json = json.dumps(self.get_jwk(), sort_keys=True, separators=(",", ":"))
+        acme_thumbprint = safe_base64(sha256(acme_header_jwk_json.encode("utf8")).digest())
+        acme_keyauthorization = "{0}.{1}".format(token, acme_thumbprint)
 
-        return acme_keyauthorization, base64_of_acme_keyauthorization
+        return acme_keyauthorization
 
     def check_authorization_status(self, authorization_url, desired_status=None):
         """
@@ -466,16 +516,15 @@ class Client(object):
         number_of_checks = 0
         while True:
             time.sleep(self.ACME_AUTH_STATUS_WAIT_PERIOD)
-            headers = {"User-Agent": self.User_Agent}
-            check_authorization_status_response = requests.get(
-                authorization_url, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers
+            check_authorization_status_response = self.make_signed_acme_request(
+                authorization_url, payload=""
             )
             authorization_status = check_authorization_status_response.json()["status"]
             number_of_checks = number_of_checks + 1
             self.logger.debug(
                 "check_authorization_status_response. status_code={0}. response={1}".format(
                     check_authorization_status_response.status_code,
-                    self.log_response(check_authorization_status_response),
+                    log_response(check_authorization_status_response),
                 )
             )
             if authorization_status in desired_status:
@@ -492,7 +541,7 @@ class Client(object):
         self.logger.info("check_authorization_status_success")
         return check_authorization_status_response
 
-    def respond_to_challenge(self, acme_keyauthorization, dns_challenge_url):
+    def respond_to_challenge(self, acme_keyauthorization, challenge_url):
         """
         https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.5.1
         To prove control of the identifier and receive authorization, the
@@ -507,13 +556,15 @@ class Client(object):
         request to the authorization URL, and the server responds with the
         current authorization object.
         """
-        self.logger.info("respond_to_challenge")
-        payload = {"keyAuthorization": "{0}".format(acme_keyauthorization)}
-        respond_to_challenge_response = self.make_signed_acme_request(dns_challenge_url, payload)
+        self.logger.info(
+            "respond_to_challenge for %s at %s" % (acme_keyauthorization, challenge_url)
+        )
+        payload = json.dumps({"keyAuthorization": "{0}".format(acme_keyauthorization)})
+        respond_to_challenge_response = self.make_signed_acme_request(challenge_url, payload)
         self.logger.debug(
             "respond_to_challenge_response. status_code={0}. response={1}".format(
                 respond_to_challenge_response.status_code,
-                self.log_response(respond_to_challenge_response),
+                log_response(respond_to_challenge_response),
             )
         )
 
@@ -534,11 +585,13 @@ class Client(object):
         GET request to the order resource to obtain its current state.
         """
         self.logger.info("send_csr")
-        payload = {"csr": self.calculate_safe_base64(self.csr)}
-        send_csr_response = self.make_signed_acme_request(url=finalize_url, payload=payload)
+        payload = {"csr": safe_base64(self.csr)}
+        send_csr_response = self.make_signed_acme_request(
+            url=finalize_url, payload=json.dumps(payload)
+        )
         self.logger.debug(
             "send_csr_response. status_code={0}. response={1}".format(
-                send_csr_response.status_code, self.log_response(send_csr_response)
+                send_csr_response.status_code, log_response(send_csr_response)
             )
         )
 
@@ -546,7 +599,7 @@ class Client(object):
             raise ValueError(
                 "Error sending csr: status_code={status_code} response={response}".format(
                     status_code=send_csr_response.status_code,
-                    response=self.log_response(send_csr_response),
+                    response=log_response(send_csr_response),
                 )
             )
         send_csr_response_json = send_csr_response.json()
@@ -555,29 +608,22 @@ class Client(object):
         self.logger.info("send_csr_success")
         return certificate_url
 
-    def download_certificate(self, certificate_url):
+    def download_certificate(self, certificate_url: str) -> str:
         self.logger.info("download_certificate")
 
-        download_certificate_response = self.make_signed_acme_request(
-            certificate_url, payload="DOWNLOAD_Z_CERTIFICATE"
-        )
+        response = self.make_signed_acme_request(certificate_url, payload="")
         self.logger.debug(
             "download_certificate_response. status_code={0}. response={1}".format(
-                download_certificate_response.status_code,
-                self.log_response(download_certificate_response),
+                response.status_code, log_response(response)
             )
         )
-
-        if download_certificate_response.status_code not in [200, 201]:
+        if response.status_code not in [200, 201]:
             raise ValueError(
                 "Error fetching signed certificate: status_code={status_code} response={response}".format(
-                    status_code=download_certificate_response.status_code,
-                    response=self.log_response(download_certificate_response),
+                    status_code=response.status_code, response=log_response(response)
                 )
             )
-
-        pem_certificate = download_certificate_response.content.decode("utf-8")
-
+        pem_certificate = response.content.decode("utf-8")
         self.logger.info("download_certificate_success")
         return pem_certificate
 
@@ -593,10 +639,7 @@ class Client(object):
         in order to protect against replay attacks.
         """
         self.logger.debug("get_nonce")
-        headers = {"User-Agent": self.User_Agent}
-        response = requests.get(
-            self.ACME_GET_NONCE_URL, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers
-        )
+        response = self.GET(self.ACME_GET_NONCE_URL)
         nonce = response.headers["Replay-Nonce"]
         return nonce
 
@@ -618,18 +661,29 @@ class Client(object):
             payload[k] = v
         return payload
 
-    @staticmethod
-    def calculate_safe_base64(un_encoded_data):
+    def get_jwk(self):
         """
-        takes in a string or bytes
-        returns a string
+        calculate the JSON Web Key (jwk) from self.account_key
         """
-        if isinstance(un_encoded_data, str):
-            un_encoded_data = un_encoded_data.encode("utf8")
-        r = base64.urlsafe_b64encode(un_encoded_data).rstrip(b"=")
-        return r.decode("utf8")
+        private_key = cryptography.hazmat.primitives.serialization.load_pem_private_key(
+            self.account_key.encode(),
+            password=None,
+            backend=cryptography.hazmat.backends.default_backend(),
+        )
+        public_key_public_numbers = private_key.public_key().public_numbers()
+        # private key public exponent in hex format
+        exponent = "{0:x}".format(public_key_public_numbers.e)
+        exponent = "0{0}".format(exponent) if len(exponent) % 2 else exponent
+        # private key modulus in hex format
+        modulus = "{0:x}".format(public_key_public_numbers.n)
+        jwk = {
+            "kty": "RSA",
+            "e": safe_base64(binascii.unhexlify(exponent)),
+            "n": safe_base64(binascii.unhexlify(modulus)),
+        }
+        return jwk
 
-    def get_acme_header(self, url):
+    def get_acme_header(self, url, needs_jwk=False):
         """
         https://tools.ietf.org/html/draft-ietf-acme-acme#section-6.2
         The JWS Protected Header MUST include the following fields:
@@ -641,106 +695,139 @@ class Client(object):
         """
         self.logger.debug("get_acme_header")
         header = {"alg": "RS256", "nonce": self.get_nonce(), "url": url}
-        if url in [self.ACME_NEW_ACCOUNT_URL, self.ACME_REVOKE_CERT_URL, "GET_THUMBPRINT"]:
-            private_key = cryptography.hazmat.primitives.serialization.load_pem_private_key(
-                self.account_key.encode(),
-                password=None,
-                backend=cryptography.hazmat.backends.default_backend(),
-            )
-            public_key_public_numbers = private_key.public_key().public_numbers()
-            # private key public exponent in hex format
-            exponent = "{0:x}".format(public_key_public_numbers.e)
-            exponent = "0{0}".format(exponent) if len(exponent) % 2 else exponent
-            # private key modulus in hex format
-            modulus = "{0:x}".format(public_key_public_numbers.n)
-            jwk = {
-                "kty": "RSA",
-                "e": self.calculate_safe_base64(binascii.unhexlify(exponent)),
-                "n": self.calculate_safe_base64(binascii.unhexlify(modulus)),
-            }
-            header["jwk"] = jwk
+
+        if needs_jwk:
+            header["jwk"] = self.get_jwk()
         else:
             header["kid"] = self.kid
         return header
 
-    def make_signed_acme_request(self, url, payload):
+    def make_signed_acme_request(self, url, payload, needs_jwk=False):
         self.logger.debug("make_signed_acme_request")
-        headers = {"User-Agent": self.User_Agent}
-        payload = self.stringfy_items(payload)
-
-        if payload in ["GET_Z_CHALLENGE", "DOWNLOAD_Z_CERTIFICATE"]:
-            response = requests.get(url, timeout=self.ACME_REQUEST_TIMEOUT, headers=headers)
-        else:
-            payload64 = self.calculate_safe_base64(json.dumps(payload))
-            protected = self.get_acme_header(url)
-            protected64 = self.calculate_safe_base64(json.dumps(protected))
-            signature = self.sign_message(message="{0}.{1}".format(protected64, payload64))  # bytes
-            signature64 = self.calculate_safe_base64(signature)  # str
-            data = json.dumps(
-                {"protected": protected64, "payload": payload64, "signature": signature64}
-            )
-            headers.update({"Content-Type": "application/jose+json"})
-            response = requests.post(
-                url, data=data.encode("utf8"), timeout=self.ACME_REQUEST_TIMEOUT, headers=headers
-            )
+        headers = {}
+        payload64 = safe_base64(payload)
+        protected = self.get_acme_header(url, needs_jwk)
+        protected64 = safe_base64(json.dumps(protected))
+        signature = self.sign_message(message="{0}.{1}".format(protected64, payload64))  # bytes
+        signature64 = safe_base64(signature)  # str
+        data = json.dumps(
+            {"protected": protected64, "payload": payload64, "signature": signature64}
+        )
+        headers.update({"Content-Type": "application/jose+json"})
+        response = self.POST(url, data=data.encode("utf8"), headers=headers)
         return response
 
     def get_certificate(self):
         self.logger.debug("get_certificate")
-        domain_dns_value = "placeholder"
-        dns_names_to_delete = []
+        challenges = []
+
         try:
             self.acme_register()
             authorizations, finalize_url = self.apply_for_cert_issuance()
-            responders = []
-            for url in authorizations:
-                identifier_auth = self.get_identifier_authorization(url)
-                authorization_url = identifier_auth["url"]
-                dns_name = identifier_auth["domain"]
-                dns_token = identifier_auth["dns_token"]
-                dns_challenge_url = identifier_auth["dns_challenge_url"]
 
-                acme_keyauthorization, domain_dns_value = self.get_keyauthorization(dns_token)
-                self.dns_class.create_dns_record(dns_name, domain_dns_value)
-                dns_names_to_delete.append(
-                    {"dns_name": dns_name, "domain_dns_value": domain_dns_value}
-                )
-                responders.append(
-                    {
-                        "authorization_url": authorization_url,
-                        "acme_keyauthorization": acme_keyauthorization,
-                        "dns_challenge_url": dns_challenge_url,
-                    }
-                )
+            for auth_url in authorizations:
+                identifier_auth = self.get_identifier_authorization(auth_url)
+                token = identifier_auth["token"]
+                challenge = {
+                    "ident_value": identifier_auth["domain"],
+                    "token": token,
+                    "key_auth": self.get_keyauthorization(token),  # responder acme_keyauth..
+                    "wildcard": identifier_auth["wildcard"],
+                    "auth_url": auth_url,  # responder auth.._url
+                    "chal_url": identifier_auth["challenge_url"],  # responder challenge_url
+                }
+                challenges.append(challenge)
+
+            # any errors in setup are fatal (here - they are all necessary for same cert)
+            failures = self.provider.setup(challenges)
+            if failures:
+                raise RuntimeError("get_certificate: challenge setup failed for %s" % failures)
+
+            ### FIX ME ### should abort cert and try to clear on error
+
+            self.propagation_delay(challenges)
 
             # for a case where you want certificates for *.example.com and example.com
-            # you have to create both dns records AND then respond to the challenge.
+            # you have to create both auth records AND then respond to the challenge.
             # see issues/83
-            for i in responders:
+            for chal in challenges:
                 # Make sure the authorization is in a status where we can submit a challenge
                 # response. The authorization can be in the "valid" state before submitting
                 # a challenge response if there was a previous authorization for these hosts
                 # that was successfully validated, still cached by the server.
-                auth_status_response = self.check_authorization_status(i["authorization_url"])
+                auth_status_response = self.check_authorization_status(chal["auth_url"])
                 if auth_status_response.json()["status"] == "pending":
-                    self.respond_to_challenge(i["acme_keyauthorization"], i["dns_challenge_url"])
+                    self.respond_to_challenge(chal["key_auth"], chal["chal_url"])
 
-            for i in responders:
+            ### TO DO ### this is the obfuscated timeout loop.  Clean this mess up!
+            ### # # # ### it also keeps trying even when the auth is failed :-(
+
+            for chal in challenges:
                 # Before sending a CSR, we need to make sure the server has completed the
                 # validation for all the authorizations
-                self.check_authorization_status(i["authorization_url"], ["valid"])
+                self.check_authorization_status(chal["auth_url"], ["valid"])
 
             certificate_url = self.send_csr(finalize_url)
             certificate = self.download_certificate(certificate_url)
+
+        ### FIX ME ### [:100] is a bandaid to reduce spew during tests
+
         except Exception as e:
-            self.logger.error("Error: Unable to issue certificate. error={0}".format(str(e)))
+            self.logger.error("Error: Unable to issue certificate. error={0}".format(str(e)[:100]))
             raise e
         finally:
-            for i in dns_names_to_delete:
-                self.dns_class.delete_dns_record(i["dns_name"], i["domain_dns_value"])
+            # best-effort attempt to clear challenges
+            failures = self.provider.clear(challenges)
 
         return certificate
 
+    def sleep_iter(self):
+        "returns values from list, then repeats last value forever"
+
+        for cur_time in self.provider.prop_sleep_times:
+            yield cur_time
+        while True:
+            yield cur_time
+
+    def propagation_delay(self, challenges):
+        """
+        Wait for the challenges to propagate through the service.
+
+        SHOULD return success or failure... TBD.
+
+        This process is controlled by two attributes and one method on the provider.
+        unpropagated() is called to check whether the listed challenges are thought
+        to be ready to be validated by the ACME server.  If it fails, propagation_delay
+        will sleep and retry until it times out.  The attributes on the provider are
+        prop_timeout and prop_sleep_times.  See docs/unpropagated.md for the details.
+        """
+
+        if self.provider.prop_timeout:
+            unready = challenges
+            end_time = time.time() + self.provider.prop_timeout
+            next_sleep = self.sleep_iter()
+            num_checks = 0
+            while unready:
+                errata = self.provider.unpropagated(unready)
+                num_checks += 1
+
+                ### FIX ME ### if any errata have status "failed", exit with error
+                if errata:
+                    poll_time = time.time()
+                    # intentional: do an "extra" check rather than running short
+                    if end_time < poll_time:
+                        break
+                    # wait a while to let more propagation happen
+                    time.sleep(next(next_sleep))
+
+                unready = [err[2] for err in errata]
+
+            ### FIX ME ### might be good for mock tests, but really should try to clear, eh?
+            if unready:
+                raise RuntimeError(
+                    "propagation_delay: time out after %s probes: %s" % (num_checks, unready)
+                )
+
     def cert(self):
         """
         convenience method to get a certificate without much hassle
