@@ -1,4 +1,5 @@
-from .common import BaseDns  # noqa: F401
+### legacy DNS bindings - DEPRECATED ###
+
 from .auroradns import AuroraDns  # noqa: F401
 from .cloudflare import CloudFlareDns  # noqa: F401
 from .acmedns import AcmeDnsDns  # noqa: F401
@@ -9,3 +10,5 @@ from .dnspod import DNSPodDns  # noqa: F401
 from .duckdns import DuckDNSDns  # noqa: F401
 from .cloudns import ClouDNSDns  # noqa: F401
 from .route53 import Route53Dns  # noqa: F401
+from .powerdns import PowerDNSDns
+from .gandi import GandiDns
