@@ -1,31 +1,42 @@
-import logging
+from hashlib import sha256
+from typing import Any, Dict, Sequence
+
+from sewer.auth import ErrataItemType, ProviderBase
+from sewer.lib import safe_base64
 
 
-class BaseDns(object):
+def dns_challenge(key_auth: str) -> str:
+    "return safe-base64 of hash of key_auth; used for dns response"
+
+    return safe_base64(sha256(key_auth.encode("utf8")).digest())
+
+
+class BaseDns(ProviderBase):
     """
+    Shim for legacy DNS provider interface.
     """
 
-    def __init__(self, LOG_LEVEL="INFO"):
-        self.LOG_LEVEL = LOG_LEVEL
-        self.dns_provider_name = self.__class__.__name__
+    def __init__(self, **kwargs: Any) -> None:
+        if "chal_types" not in kwargs:
+            kwargs["chal_types"] = ["dns-01"]
+        super().__init__(**kwargs)
 
-        self.logger = logging.getLogger(__name__)
-        self.logger.setLevel(self.LOG_LEVEL)
-        if not self.logger.hasHandlers():
-            handler = logging.StreamHandler()
-            formatter = logging.Formatter("%(message)s")
-            handler.setFormatter(formatter)
-            self.logger.addHandler(handler)
+    ### shim methods
 
-    def log_response(self, response):
-        """
-        renders a python-requests response as json or as a string
-        """
-        try:
-            log_body = response.json()
-        except ValueError:
-            log_body = response.content
-        return log_body
+    def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
+        for chal in challenges:
+            self.create_dns_record(chal["ident_value"], dns_challenge(chal["key_auth"]))
+        return []
+
+    def unpropagated(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
+        return []
+
+    def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
+        for chal in challenges:
+            self.delete_dns_record(chal["ident_value"], dns_challenge(chal["key_auth"]))
+        return []
+
+    ### legacy DNS methods
 
     def create_dns_record(self, domain_name, domain_dns_value):
         """
@@ -58,7 +69,6 @@ class BaseDns(object):
             Please consult your dns provider on how/format of their DNS TXT records.
             You may also want to consult the cloudflare DNS implementation that is found in this repository.
         """
-        self.logger.info("create_dns_record")
         raise NotImplementedError("create_dns_record method must be implemented.")
 
     def delete_dns_record(self, domain_name, domain_dns_value):
@@ -73,5 +83,4 @@ class BaseDns(object):
 
         This method should return None
         """
-        self.logger.info("delete_dns_record")
         raise NotImplementedError("delete_dns_record method must be implemented.")
